import requests
import json
from .SpeedyCard import SpeedyCard
from typing import NamedTuple


class MessageResponse(NamedTuple):
    id: str
    roomId: str
    toPersonEmail: str
    roomType: str
    text: str
    personId: str
    personEmail: str
    markdown: str
    html: str
    created: str


class SpeedyBot:
    def __init__(self, token: str, make_request=requests.request):
        self.token = token
        self.middlewares = []
        self.top_middleware = None
        self.reject_middleware = None
        self.make_request = make_request
        self.API = {
            "messages": "https://webexapis.com/v1/messages",
            "attachments": "https://webexapis.com/v1/attachment/actions",
            "user": {
                "self": "https://webexapis.com/v1/people/me",
                "get_person_details": "https://webexapis.com/v1/people",
            },
            "rooms": "https://developer.webex.com/docs/api/v1/rooms/list-rooms",
            "room_details": "https://webexapis.com/v1/rooms",
            "webhooks": "https://webexapis.com/v1/webhooks",
        }

    def set_token(self, token):
        self.token = token

    def card(self):
        return SpeedyCard()

    def send_card_to(
        self,
        email_or_room_id: str,
        card: SpeedyCard,
        fallback="Your client does not support adaptive cards",
    ) -> MessageResponse:
        headers = {
            "Authorization": f"Bearer {self.token}",
            "Content-Type": "application/json",
        }

        payload = {
            "toPersonEmail": email_or_room_id if "@" in email_or_room_id else None,
            "roomId": email_or_room_id if "@" not in email_or_room_id else None,
            "markdown": fallback,
            "attachments": [
                {
                    "contentType": "application/vnd.microsoft.card.adaptive",
                    "content": card.build(),
                }
            ],
        }
        response = self.make_request(
            "POST", self.API["messages"], headers=headers, data=json.dumps(payload)
        )
        return response.json()

    def send_to(self, email_or_room_id, message) -> MessageResponse:
        """
        Send a message to an email or room. The message can be either a string or a SpeedyCard

        :param email_or_room_id: email address OR room id
        :param message: The message content to reply with.
        """

        if isinstance(message, SpeedyCard):
            return self.send_card_to(email_or_room_id, message)

        headers = {
            "Authorization": f"Bearer {self.token}",
            "Content-Type": "application/json",
        }
        payload = {"toPersonEmail": email_or_room_id, "markdown": message}
        response = self.make_request(
            "POST", self.API["messages"], headers=headers, data=json.dumps(payload)
        )
        return response.json()

    def reply(self, roomId, messageId, message: str) -> MessageResponse:
        """
        Reply to a message in a chat room.

        :param roomId: The ID of the chat room (not name)
        :param messageId: The ID of the message to reply, this message must exist or else the request will fail
        :param message: The message content to reply with.
        :return: The result of the reply operation as a string.
        """
        headers = {
            "Authorization": f"Bearer {self.token}",
            "Content-Type": "application/json",
        }
        payload = {
            "parentId": messageId,
            "roomId": roomId,
            "text": message,
            "markdown": message,
        }
        response = self.make_request(
            "POST", self.API["messages"], headers=headers, data=json.dumps(payload)
        )
        return response.json()

    ## will unlock this later
    ## see https://speedybot.js.org to roll your own (not hard, just bunch of steps)
    # def add_middleware(self, middleware):
    #     self.middlewares.append(middleware)

    # def run_middleware(self, ctx):
    #     if self.top_middleware:
    #         result = self.top_middleware(ctx)
    #         if not result and self.reject_middleware:
    #             self.reject_middleware(ctx)
    #             return False
    #     for middleware in self.middlewares:
    #         result = middleware(ctx)
    #         if not result:
    #             if self.reject_middleware:
    #                 self.reject_middleware(ctx)
    #             break
    #     return True



import copy


class SpeedyCard:
    def __init__(self):
        self.json = {
            "$schema": "http://adaptivecards.io/schemas/adaptive-card.json",
            "type": "AdaptiveCard",
            "version": "1.0",
            "body": [],
        }
        self._stash = {
            "needsSubmit": False,
            "title": "",
            "subTitle": "",
            "chips": [],
            "data": {},
            "submitLabel": "Submit",
            "backgroundImage": "",
        }

        self.id = {}

    def check_id(self, id=""):
        if id in self.id:
            self.id[id] += 1
            return f"{id}_{self.id[id]}"
        else:
            self.id[id] = 1
            return id

    def add_title(self, title: str):
        """
        Args:
            title (str): card's title
        """

        self._stash["title"] = title
        return self

    def add_subtitle(self, subTitle: str):
        """
        Args:
            subTitle (str): card's subtitle
        """
        self._stash["subTitle"] = subTitle
        return self

    def add_image(self, url, horizontalAlignment="Center", size="ExtraLarge"):
        """
        Arguments are case-sensitive
        Args:
            horizontalAlignment (str): Horizontal alignment ("Left", "Center", "Right").
            size (str): Size option ("Small", "Default", "Medium", "Large", "ExtraLarge").
        """
        if url:
            self.json["body"].append(
                {
                    "horizontalAlignment": horizontalAlignment,
                    "size": size,
                    "type": "Image",
                    "url": url,
                }
            )
        return self

    def add_subcard(self, card: "SpeedyCard", label="Open"):
        """
        Args:
            card (SpeedyCard): adds a subcard, must be a SpeedyCard
        """
        payload = {"type": "Action.ShowCard", "title": label, "card": card.build()}
        self.add_action(payload)
        return self

    def add_text(self, text: str, horizontalAlignment=None, size=None, color=None):
        """
        Arguments are case-sensitive
        Args:
            horizontalAlignment (str): Horizontal alignment ("Left", "Center", "Right").
            size (str): Size option ("Small", "Default", "Medium", "Large", "ExtraLarge").
            color (str): ("Default", "Dark", "Light", "Accent", "Good", "Warning", "Attention").
        """

        self.json["body"].append(
            {
                "type": "TextBlock",
                "text": text,
                "wrap": True,
                "size": size,
                "horizontalAlignment": horizontalAlignment,
                "color": color,
            }
        )
        return self

    def add_table(self, input, separator=False):
        if not isinstance(input, list) or not all(
            isinstance(item, list) for item in input
        ):
            raise ValueError("Input should be a list of lists")

        facts = [{"title": item[0], "value": item[1]} for item in input]
        result = {"type": "FactSet", "separator": separator, "facts": facts}
        self.json["body"].append(result)
        return self

    def buildTextPayload(self, text, config):
        if config is None:
            config = {}

        size = config.get("size", "Medium")
        align = config.get("align", "Left")
        color = config.get("color")

        text_block = {
            "type": "TextBlock",
            "text": text,
            "wrap": True,
            "size": size,
            "horizontalAlignment": align,
            **({"color": color} if color is not None else {}),
        }
        return text_block

    def setBackgroundImage(self, url: str):
        self._stash["backgroundImage"] = url
        return self

    def add_link(self, url: str, label: str = None):
        if label is None:
            label = url
        self.add_text(f"**[{label}]({url})**")
        return self

        self.json["body"].append(
            {
                type: "Image",
                "size": size,
                "horizontalAlignment": horizontalAlignment,
                url: url,
            }
        )
        return self

    def build(self):
        result_json = copy.deepcopy(self.json)
        if self._stash["subTitle"]:
            result_json["body"].insert(
                0, self.buildTextPayload(self._stash["subTitle"], {})
            )

        if self._stash["title"]:
            result_json["body"].insert(
                0,
                self.buildTextPayload(
                    self._stash["title"],
                    {
                        "weight": "Bolder",
                        "size": "ExtraLarge",
                    },
                ),
            )

        if self._stash["backgroundImage"]:
            result_json["backgroundImage"] = self._stash["backgroundImage"]

        # if self._stash["needsSubmit"]:
        #     payload = {"type": "Action.Submit", "title": self._stash["submitLabel"]}

        return result_json

    # Stay tuned, if you need something now, see https://speedybot.js.org
    def add_action(self, action):
        if "actions" not in self.json:
            self.json["actions"] = []
        self.json["actions"].append(action)
